// LeftistHeap class
//
// CONSTRUCTION: with a negative infinity sentinel
//
// ******************PUBLIC OPERATIONS*********************
// void insert( x )       --> Insert x
// Comparable deleteMin( )--> Return and remove smallest item
// Comparable findMin( )  --> Return smallest item
// boolean isEmpty( )     --> Return true if empty; else false
// boolean isFull( )      --> Return false in this implementation
// void makeEmpty( )      --> Remove all items
// void merge( rhs )      --> Absorb rhs into this heap

/**
 * Implements a leftist heap.
 * Note that all "matching" is based on the compareTo method.
 *
 * @author Mark Allen Weiss
 */
public class LeftistHeap {
   /**
    * Construct the leftist heap.
    */
   public LeftistHeap() {
      root = null;
   }

   /**
    * Merge rhs into the priority queue.
    * rhs becomes empty. rhs must be different from this.
    *
    * @param rhs the other leftist heap.
    */
   public void merge(LeftistHeap rhs) {
      if (this == rhs)    // Avoid aliasing problems
         return;

      root = merge(root, rhs.root);
      rhs.root = null;
   }

   /**
    * Insert into the priority queue, maintaining heap order.
    *
    * @param x the item to insert.
    */
   public void insert(Comparable x) {
      root = merge(new LeftHeapNode(x), root);
   }

   /**
    * Remove the smallest item from the priority queue.
    *
    * @return the smallest item, or null, if empty.
    */
   public Comparable deleteMin() {
      if (isEmpty())
         return null;

      Comparable minItem = root.element;
      root = merge(root.left, root.right);

      return minItem;
   }

   private LeftHeapNode root;    // root

   @Override
   public String toString() {
      if(isEmpty())
         return "Empty heap!";
      else
         return root.toString();
   }

   public String nplStr() {
      if(isEmpty())
         return "Empty NPL";
      else
         return root.nplStr();
   }

   /**
    * Test if the priority queue is logically empty.
    *
    * @return true if empty, false otherwise.
    */
   private boolean isEmpty() {
      return root == null;
   }

   /**
    * Internal static method to merge two roots.
    * Deals with deviant cases and calls recursive merge1.
    *
    * @param h1 The first root.
    * @param h2 The second root.
    * @return The new merged root.
    */
   private static LeftHeapNode merge(LeftHeapNode h1, LeftHeapNode h2) {
      if (h1 == null)
         return h2;
      if (h2 == null)
         return h1;
      if (h1.element.compareTo(h2.element) < 0)
         return merge1(h1, h2);
      else
         return merge1(h2, h1);
   }

   /**
    * Internal static method to merge two roots.
    * Assumes trees are not empty, and h1's root contains smallest item.
    *
    * @param h1 The first root.
    * @param h2 The second root.
    * @return The new merged root.
    */
   private static LeftHeapNode merge1(LeftHeapNode h1, LeftHeapNode h2) {
      if (h1.left == null)   // Single node
         h1.left = h2;       // Other fields in h1 already accurate
      else {
         // Since we are not switching nodes we merge with the current NPL
         if(h1.getNPLNode() == h1.right) {
            h1.right = merge(h1.right, h2);
         } else {
            h1.left = merge(h1.left, h2);
	 }
	 h1.npl = Math.min(h1.right.npl,h1.left.npl) + 1;

      }
      return h1;
   }
}
