#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""Solve the N-puzzle game.

The idea is to use an A*-search with a number of heuristics. The simplest to
try are Manhattan distance and number of misplaced squares. Quick googling
suggests other strategies (e.g. number of swaps between any pair of tiles one
would want), but I am not certain how promising they are (for swapping
heuristic a swapping of just one pair could happen on opposite ends of the
board and is probably NOT the best guideline for the search)
"""
import math
import sys
import copy




#
# Sentinel to mark which tile on the board is free. 
#
HOLE = 0
# Directionality of tile movement.
LEFT = 'V'
RIGHT = 'H'
UP = 'O'
DOWN = 'N'



def is_solvable(board):
    """Exercise 23.2 (formula 23.3.4)

    This checks whether a given board is solvable, if it is an initial
    configuration.
    """

    N = board.N
    config = board.config
    hole_index = config.index(HOLE)

    def L(k):
        """Number of tiles t < k, such that the position of t comes after k
        config.
        """

        result = 0
        assert 1 <= k <= N*N
        for tile in range(1, k):
            idx = config.index(tile)
            if idx > k:
                result += 1

        return result
    # end L(k)

    sum_of_L = sum(L(k) for k in range(1, N*N+1))
    
    hole_row = (hole_index / N) + 1
    hole_column = (hole_index % N) + 1

    return sum_of_L % 2 == (hole_row + hole_column) % 2
# end is_solvable



class Board(object):
    """Class representing a board configuration in N-puzzle.

    Each configuration is stored as a tuple in row-major order.
    """

    def __init__(self, configuration, parent=None, move_count=0, directionality=None):
        """Initialiase a new board.

        @param configuration: config this board corresponds to

        @parent: a board from which this one was generated in one move. 
        """

        self.N = int(math.sqrt(len(configuration)))
        self.config = tuple(configuration)

        # This bits are for the A*-search, rather than a particular tile
        # configuration.
        # Board that self was generated from.
        self.parent = parent
        # number of moves from start. This is the actual cost of reaching this
        # configuration (through a path of parent-pointers).
        self.move_count = move_count
        # which way we moved from parent to self. Strictly speaking this is
        # not necessary: given two boards 1 move apart, we can *ALWAYS* figure
        # out which way the tile moved. However, I am lazy. Lazyness pays off
        # immediately, as does passing the directionality as a param and using
        # it in output.
        self.directionality = directionality
    # end __init__


    def __hash__(self):
        return hash(self.config)
    # end __hash__
    

    def __eq__(self, other):
        assert isinstance(other, self.__class__)
        return self.config == other.config
    # end __eq__


    def __ne__(self, other):
        # ARGH! When defining eq, do NOT forget ne. They *MUST* be
        # symmetricial (by default they are id()-based, which leads to
        # eh... interesting results))
        return not self.__eq__(other)
    # end __ne__


    def __lt__(self, other):
        # We just need *some* ordering
        return self.config < other.config
    # end __lt__
    

    def output(self, stream=sys.stdout):
        for row in range(self.N):
            stream.write(" ".join("%d" % self.config[x]
                                  for x in range(row*self.N, (row+1)*self.N)))
            stream.write("\n")
    # end output
    

    def next_move(self):
        """Yield a sequence of next valid moves from this configuration"""

        index = self.config.index(HOLE)
        N = self.N

        # move the hole down one row
        if index + N < len(self.config):
            t = list(self.config)
            t[index], t[index+N] = t[index+N], t[index]
            yield Board(t, self, self.move_count+1, DOWN)

        # move the hole up one row
        if index - N >= 0:
            t = list(self.config)
            t[index], t[index-N] = t[index-N], t[index]
            yield Board(t, self, self.move_count+1, UP)

        # move the hole left one column
        if index % N != 0:
            t = list(self.config)
            t[index], t[index-1] = t[index-1], t[index]
            yield Board(t, self, self.move_count+1, LEFT)
        
        # move the hole right one column
        if index % N != N-1:
            t = list(self.config)
            t[index], t[index+1] = t[index+1], t[index]
            yield Board(t, self, self.move_count+1, RIGHT)
    # end next_move
# end Board


class WorkQueue(object):
    """Priority queue for A*-search.

    The heap itself contains pairs -- (priority, config).
    """

    def __init__(self):
        self.q = list()

        # configuration -> its index in q. This must be updated in lock-step
        # with self.q
        self.in_queue = dict()
        self.decrease_key_count = 0
        self.push_count = 0
        self.pop_count = 0
    # end __init__


    def pop(self):
        """Stolen from heapq.heappop."""
        heap = self.q
        lastelt = heap.pop()
        if heap:
            returnitem = heap[0]
            heap[0] = lastelt
            self.in_queue[lastelt[1]] = 0
            self._siftup(0)
        else:
            returnitem = lastelt

        priority, configuration = returnitem
        del self.in_queue[configuration]
        self.pop_count += 1
        return configuration
    # end pop


    def _siftup(self, pos):
        """Stolen from heapq."""
        heap = self.q
        endpos = len(heap)
        startpos = pos
        newitem = heap[pos]
        # Bubble up the smaller child until hitting a leaf.
        childpos = 2*pos + 1    # leftmost child position
        while childpos < endpos:
            # Set childpos to index of smaller child.
            rightpos = childpos + 1
            if rightpos < endpos and not heap[childpos] < heap[rightpos]:
                childpos = rightpos
                
            # Move the smaller child up.
            heap[pos] = heap[childpos]
            # Adjust the index for the configuration in heap[childpos] (it's
            # at index pos now, not childpos)
            self.in_queue[heap[childpos][1]] = pos
            pos = childpos
            childpos = 2*pos + 1
            
        # The leaf at pos is empty now.  Put newitem there, and bubble it up
        # to its final resting place (by sifting its parents down).
        heap[pos] = newitem
        self.in_queue[newitem[1]] = pos
        self._siftdown(startpos, pos)
    # end _siftup


    def _siftdown(self, startpos, pos):
        """This is a copy of heapq._siftdown.

        However, it also keeps track of changes and updates self.in_queue,
        which keeps track of which index in the heap a given configuration is
        at.
        """

        heap = self.q
        newitem = heap[pos]
        # Follow the path to the root, moving parents down until finding a
        # place newitem fits.
        while pos > startpos:
            parentpos = (pos - 1) >> 1
            parent = heap[parentpos]
            # Although 'newitem' and 'parent' are tuples, their first element
            # is priority. That will decide the actual order and comparing
            # with < makes sense.
            if newitem < parent:
                heap[pos] = parent
                self.in_queue[parent[1]] = pos
                pos = parentpos
                continue
            break

        heap[pos] = newitem
        self.in_queue[newitem[1]] = pos
    # end _siftdown


    def decrease_key(self, position):
        self.decrease_key_count += 1
        return self._siftdown(0, position)
    # end decrease_key


    def push(self, priority, configuration):
        # we already have this configuration in the queue pending extraction.
        if configuration in self.in_queue:
            old_index = self.in_queue[configuration]
            old_priority, old_config = self.q[old_index]

            # current enqueued configuration has a no larger cost than the new
            # configuration. We can gladly discard configuration (i.e. it is
            # NOT an improvement over what is in the queue already)
            if old_priority <= priority:
                return

            # Ah, we have an actual improvement on an existing
            # configuration. Now is time to run decrease_key.
            # adjust priority
            self.q[old_index] = (priority, old_config)
            # restore heap property
            self.decrease_key(old_index)
        # it's an entirely new configuration.
        else:
            self.q.append((priority, configuration))
            self._siftdown(0, len(self.q)-1)
            self.push_count += 1
    # end push

    def __len__(self):
        return len(self.q)
    # end __len__
# end WorkQueue



def astar(heuristic, initial, goal, monotone=True):
    """Use A*-search to reach goal from initial by using heuristic.

    @type heuristic: any callable
    @param heuristic:
      Heuristic function to use.

    @type initial: A Board instance
    @param initial:
      Initial board state.

    @type goal: A Board instance.
    @param goal:
      Goal instance we want to reach.

    @param monotone:
      True iff heuristic is monotone. IT IS CRITICAL that for non-monotone
      heuristics this argument is False (otherwise we risk passing suboptimal
      solution as optimal)
    """

    # nodes under consideration
    pq = WorkQueue()
    # configurations we have seen thus far. For a monotone heuristic, once a
    # node is input, its cost cannot be improved. For non-monotone heuristic
    # that is not the case.
    seen = dict()
    pq.push(heuristic(initial), initial)
    this = None
    
    while pq:
        this = pq.pop()
        if this == goal:
            break

        # With a monotone heuristic this value cannot be improved
        seen[this] = this.move_count

        for child in this.next_move():
            # child.move_count is 1+this.move_count. This is how children
            # configurations are created.
            child_cost = child.move_count + heuristic(child)
            if child in seen:
                if monotone:
                    assert seen[child] <= child_cost

                if seen[child] <= child_cost:
                    continue

            pq.push(child_cost, child)

    moves = list()
    if this == goal:
        while this:
            moves.append(this)
            this = this.parent

    return moves[::-1], pq.pop_count, pq.push_count, pq.decrease_key_count
# end astar



def misfit_tiles(board):
    """Heuristic 1 - Number of tiles out of place.

    This heuristic is monotone.
    """

    return sum(int(x-1 != y)
               for y,x in enumerate(board.config)
               if x != HOLE)
# end misfit_tiles


def manhattan_distance(board):
    """Heuristic 2 - sum of Manhattan distances between all tiles and their
    respective correct places.

    This heuristic is monotone.
    """

    N = board.N
    distance = 0
    for row in range(N):
        for column in range(N):
            number = board.config[row*N + column]
            if number == HOLE:
                continue
            
            proper_row = (number-1) / N
            proper_column = (number-1) % N

            distance += abs(proper_row - row) + abs(proper_column - column)
    return distance
# end manhattan_distance



def stream_to_board(infile):
    """Construct a board configuration from infile.

    Return a pair -- initial configuration, goal configuration.

    The goal configuration is fixed for every board dimension, but it is
    conceptually easier just to have it generated with the input.
    """

    N = int(infile.readline())
    configuration = list()
    for row_number in range(N):
        configuration.extend([int(x)
                              for x in infile.readline().strip().split()])

    assert len(configuration) == N*N
    return (Board(configuration),
            Board(range(1, N*N) + [HOLE,]))
# end stream_to_board



def main(argv):
    infile = open(argv[1])
    if len(argv) < 3:
        outfile = sys.stdout
    else:
        outfile = open(argv[2], "w")

    initial, goal = stream_to_board(infile)
    print "Initial configuration is", is_solvable(initial) and "solvable" or "unsolvable"
    (moves, pop_count,
     push_count, decrease_count) = astar(manhattan_distance, initial, goal)

    outfile.write("%d\n" % (len(moves)-1))
    for move in moves:
        if move.directionality:
            outfile.write(move.directionality)
    outfile.write("\n")
    outfile.write("%d\n%d\n%d\n" % (pop_count, push_count, decrease_count))
# end main





if __name__ == "__main__":
    main(sys.argv[:])
